#include <libdragon.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

TRegBlock exception_rb;

void register_dump(void);
void reset_handler(void);
ExceptionResumeTarget exception_handler(const TExceptionBlock *eb);

void register_dump(void)
{
    static const char gprnames[32][2] = // unterminated
    {{'r','0'},{'a','t'},{'v','0'},{'v','1'},
     {'a','0'},{'a','1'},{'a','1'},{'a','3'},
     {'t','0'},{'t','1'},{'t','2'},{'t','3'},
     {'t','4'},{'t','5'},{'t','6'},{'t','7'},
     {'s','0'},{'s','1'},{'s','2'},{'s','3'},
     {'s','4'},{'s','5'},{'s','6'},{'s','7'},
     {'t','8'},{'t','9'},{'k','0'},{'k','1'},
     {'g','p'},{'s','p'},{'s','8'},{'r','a'}};

    static const char * causenames[32] =
    {"Interrupt",                   // 0
     "TLB modification",            // 1
     "TLB (load or inst)",          // 2
     "TLB (store)",                 // 3
     "Address error (load or inst)",// 4
     "Address error (store)",       // 5
     "Bus error (inst)",            // 6
     "Bus error (data)",            // 7
     "Syscall",                     // 8
     "Breakpoint",                  // 9
     "Reserved instruction",        // 10
     "Coprocessor unusable",        // 11
     "Arithmetic overflow",         // 12
     "Trap",                        // 13
     "Virtual coherency (inst)",    // 14
     "Floating point",              // 15
     "Reserved",                    // 16
     "Reserved",                    // 17
     "Reserved",                    // 18
     "Reserved",                    // 19
     "Reserved",                    // 20
     "Reserved",                    // 21
     "Reserved",                    // 22
     "Watch",                       // 23
     "Reserved",                    // 24
     "Reserved",                    // 25
     "Reserved",                    // 26
     "Reserved",                    // 27
     "Reserved",                    // 28
     "Reserved",                    // 29
     "Reserved",                    // 30
     "Virtual coherency (data)"};   // 31

    // ideally the cause register will still be valid
    uint32_t cause, badvaddr = 0;
    asm volatile ("\tmfc0 %0, $13\n"
                  "\tmfc0 %1, $8\n" : "=r" (cause), "=r" (badvaddr) : : );

    console_init();
   
    console_set_render_mode(RENDER_MANUAL);

    console_clear();
    console_printf("  %s\n", causenames[(cause >> 2) & 31]);
    console_printf("  ----------------------------\n");

    console_printf("  EPC:%08x BadAdr:%08x\n", (unsigned int)exception_rb.epc, badvaddr);
    console_printf("Cause:%08x    SR: %08x\n\n", cause, (unsigned int)exception_rb.sr);
    for (int i = 0; i < 32; i+=2)
    {
        if (i == 0)
        {
            console_printf("                  %.2s: %08lx\n", gprnames[1], exception_rb.gpr[1]);
        }
        else
        {
            console_printf("  %.2s: %08lx    %.2s: %08lx\n", gprnames[i], exception_rb.gpr[i], gprnames[i+1], exception_rb.gpr[i+1]);
        }
    }
    console_printf("  HI: %08x    LO: %08x\n", exception_rb.hi, exception_rb.lo);

    console_render();

    while (1) {}
}

void reset_handler(void)
{
    int lines_down = 0, dir = 1;
    //console_init();
   
    console_set_render_mode(RENDER_MANUAL);

    while (1)
    {
        console_clear();

#if 0
        struct controller_data data;

        controller_read(&data);

        if (data.c[0].A)
        {
            console_printf("%016llx\n", eeprom_read(0));
            console_printf("%016llx\n", eeprom_read(1));
            console_printf("%016llx\n", eeprom_read(2));
            console_printf("%016llx\n", eeprom_read(3));
        }
#endif

        for (int i = 0; i < lines_down; i++)
        {
            console_printf("\n");
        }
        console_printf("      Bye!\n");

        console_render();

        wait_ms(50);

        lines_down += dir;
        if (lines_down == 0) dir = 1;
        if (lines_down == 15) dir = -1;

    }
}

ExceptionResumeTarget exception_handler(const TExceptionBlock *eb)
{
    if (eb->type == EXCEPTION_TYPE_CRITICAL)
    {
        exception_rb = *eb->regs;
        return register_dump;
    }
    else if (eb->type == EXCEPTION_TYPE_RESET)
    {
        return reset_handler;
    }

    return NULL;
}

int
main(int argc, char *argv[])
{
    // libdragon setup
    set_MI_interrupt(1,1);
    audio_init(32000,2);
    console_init();
    console_set_render_mode(RENDER_MANUAL);
    controller_init();

    register_exception_handler( exception_handler );

    display_init(RESOLUTION_320x240, DEPTH_16_BPP, 2, GAMMA_NONE, ANTIALIAS_RESAMPLE);
    
    console_printf("\nException handler demo\n\n");
    console_printf( "A: TLB load exception\n"
                    "B: TLB store exception\n"
                    "Z: Arithmetic overflow\n"
                    "L: Break\n"
                    "R: Trap\n"
                    "D-Up:    Address error load\n"
                    "D-Down:  Address error store\n"
                    "D-Left:  Coprocessor unusable\n"
                    "D-Right: Reserved instruction\n"
                    "Start:   Syscall\n\n");
    console_printf("Press reset for \'Pre-NMI\'\n");
    console_render();

    console_set_render_mode(RENDER_AUTOMATIC);

    while (1)
    {
        struct controller_data data;

        controller_read(&data);

        if (data.c[0].B)
        {
            *(volatile int*)0 = 0;
        }
        if (data.c[0].A)
        {
            *(volatile int*)0 = *(volatile int*)0;
        }
        if (data.c[0].start)
        {
            asm volatile ("\tsyscall\n");
        }
        if (data.c[0].Z)
        {
            asm volatile(
                "\tlui  $8, 0x8000\n"
                "\taddi $8, -1\n" : : : "$8");
        }
        if (data.c[0].L)
        {
            asm volatile("\tbreak\n");
        }
        if (data.c[0].R)
        {
            asm volatile("\ttnei $0, 1\n");
        }
        if (data.c[0].down)
        {
            int target;
            *(volatile int*)((char *)(&target)+1) = 0;
        }
        if (data.c[0].up)
        {
            int target;
            *(volatile int*)0 = *(volatile int*)((char *)(&target)+1);
        }
        if (data.c[0].left)
        {
            asm volatile("\tmtc2 $0, $0\n");
        }
        if (data.c[0].right)
        {
            asm volatile("\t.dword (023 << 26)\n");
        }

        wait_ms(1);
    }
}
